<?php

declare(strict_types=1);

namespace Drupal\message_notifications\Controller;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Datetime\DateFormatter;
use Drupal\Core\Link;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\message\Entity\Message;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller class for on-site notifications.
 */
final class NotificationController extends ControllerBase {

  public const CONFIG_NAME = 'message_notifications.settings';

  /**
   * Module config.
   */
  protected ImmutableConfig $config;

  public function __construct(
    protected readonly RendererInterface $renderer,
    protected readonly DateFormatter $dateFormatter,
    ConfigFactoryInterface $configFactory,
  ) {
    $this->config = $configFactory->get(self::CONFIG_NAME);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('renderer'),
      $container->get('date.formatter'),
      $container->get('config.factory')
    );
  }

  /**
   * Get notifications block renderable.
   */
  public static function getRenderable(): array {
    return [
      '#theme' => 'message_notification_block',
      '#notifications' => [
        '#theme' => 'message_notifications',
      ],
      '#attached' => [
        'library' => [
          'message_notifications/notifications_block',
          'core/drupal.dialog.ajax',
        ],
        'drupalSettings' => [
          'notificationSystem' => [
            'getEndpointUrl' => Url::fromRoute('message_notifications.get_notifications')->toString(),
            'markAsReadEndpointUrl' => Url::fromRoute('message_notifications.markasread', [
              'message' => '0',
            ])->toString(),
          ],
        ],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getNotifications(): Response {
    $message_storage = $this->entityTypeManager()->getStorage('message');

    $notification_templates = $this->config->get('notification_templates');
    if (!\is_array($notification_templates) || \count($notification_templates) === 0) {
      return (new Response(''))->setMaxAge(0);
    }

    $display_count = $this->config->get('notification_count') ?? 5;

    $message_ids = $message_storage
      ->getQuery()
      ->accessCheck(TRUE)
      ->condition('uid', $this->currentUser()->id())
      ->condition('template', $notification_templates, 'IN')
      ->range(0, $display_count)
      ->sort('is_read', 'ASC')
      ->sort('created', 'DESC')
      ->execute();

    $messages = $message_storage->loadMultiple($message_ids);

    $build = $this->buildNotificationList($messages);
    $renderedOutput = $this->renderer->renderRoot($build);

    $response = new Response((string) $renderedOutput);
    $response->setMaxAge(0);
    return $response;
  }

  /**
   * Build notification list.
   */
  private function buildNotificationList(array $messages): array {
    $build = [];

    if (\count($messages) === 0) {
      return $build;
    }

    foreach ($messages as $message) {
      // We need to strip tags as Message getText() always adds <p> tags.
      $notification_title = \strip_tags($message->getText(NULL, 0)[0]);
      $link = Link::createFromRoute(
        $notification_title, 'entity.message.notification',
        [
          'message' => $message->id(),
        ],
        [
          'attributes' => [
            'class' => ['use-ajax', 'notification-markasread-trigger'],
            'data-dialog-type' => 'modal',
            'data-dialog-options' => Json::encode([
              'width' => '60%',
              'title' => $notification_title . ' | ' . $this->dateFormatter->format($message->getCreatedTime(), 'short'),
            ]),
            'data-message-id' => $message->id(),
          ],
        ]
      )->toString();

      $build[] = [
        '#theme' => 'message_notification_item',
        '#message' => $message,
        '#id' => $message->id(),
        '#timestamp' => $message->getCreatedTime(),
        '#title' => $link,
        '#isRead' => (bool) $message->get('is_read')->value,
      ];

    }

    return $build;
  }

  /**
   * Mark as read callback.
   */
  public function markAsRead(Message $message): JsonResponse {
    if ((bool) $message->get('is_read')->value) {
      $response_body = [
        'status' => 'error',
        'message' => 'Notification was already set as read for this user.',
      ];
      $statusCode = 400;
    }
    else {
      $message->set('is_read', TRUE)->save();
      $response_body = [
        'status' => 'success',
      ];
      $statusCode = 200;
    }

    return new JsonResponse($response_body, $statusCode);
  }

  /**
   * Notification display access.
   */
  public function messageAccess(Message $message, AccountInterface $account) {
    return AccessResult::allowedIf($message->getOwnerId() === $account->id())->addCacheableDependency($message)->cachePerUser();
  }

  /**
   * Display a message as a notification.
   */
  public function displayMessage(Message $message): array {
    return $this->entityTypeManager()->getViewBuilder('message')->view($message, 'default');
  }

}
